#ifndef BSETS_INVOL_GENERATOR_H
#define BSETS_INVOL_GENERATOR_H

#include <iostream>
#include <fstream>

#include "InvolSet.hpp"
#include "InvolWeights.hpp"
#include "Archive.hpp"
#include "InvolGeometry.hpp"
#include "GeneratorBase.hpp"

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

template<int Q>
class InvolGenerator :
        public GeneratorBase<InvolWeights<Q>, Invol<Q>::V, Invol<Q>::F, Q*Q+Q+1> {

private:
    static constexpr int F = Invol<Q>::F;
    static constexpr int V = Invol<Q>::V;

public:

    InvolGenerator() :
            GeneratorBase<InvolWeights<Q>, V, F, Q*Q+Q+1>(InvolGeometry<Q>()) {

    }

    void generate();

private:
    using GeneratorBase<InvolWeights<Q>, V, F, Q*Q+Q+1>::weights;
    using GeneratorBase<InvolWeights<Q>, V, F, Q*Q+Q+1>::gen;
    using GeneratorBase<InvolWeights<Q>, V, F, Q*Q+Q+1>::shipSet;

    // the current set of points at infinity
    PointSet<Q+1> currentInftySet{};

    // whether the origin belongs to the set
    bool hasOrigin{false};

    void addOrigin(); // only used for a starting configuration
    void addInfty(int pt); // only used for a starting configuration

    void removeInfty(int pt);

    void removeOrigin();

    virtual int shipSet();

public:

    // process arguments - they contain the points at infinity to be included, and/or the origin
    void processArgs(int arg, char **argv);

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
void InvolGenerator<Q>::generate() {
    // starting configuration was added in processArgs
    if (hasOrigin) {
        addOrigin();
    }
    for (int pt = 0; pt <= Q; ++pt) {
        if (currentInftySet.contains(pt)) {
            addInfty(pt);
        }
    }

    // recursively generate all fat points
    gen();

    // undo starting configuration
    for (int pt = Q; pt >= 0; --pt) {
        if (currentInftySet.contains(pt)) {
            removeInfty(pt);
        }
    }
    if (hasOrigin) {
        removeOrigin();
    }
}


template<int Q>
int InvolGenerator<Q>::shipSet() {
    int size{0};
    // print to stdout
    if (hasOrigin) {
        printf("+0");
        size++;
    }
    for (int pt = 0; pt <= Q; ++pt) {
        if (currentInftySet.contains(pt)) {
            printf("-%d", pt);
            size++;
        }
    }
    for (int pt = 0; pt < F; pt++) { // only fat points!
        if (weights.getSet().contains(pt)) {
            printf(":%d:%d", pt, pt + F);
            size += 2;
        }
    }
    printf("\n");
    return size;
}

template<int Q>
void InvolGenerator<Q>::addOrigin() {
    weights.addOrigin();
}

template<int Q>
void InvolGenerator<Q>::addInfty(int pt) {
    weights.addInfty(pt);
}

template<int Q>
void InvolGenerator<Q>::removeOrigin() {
    weights.removeOrigin();
}

template<int Q>
void InvolGenerator<Q>::removeInfty(int pt) {
    weights.removeInfty(pt);
}

template<int Q>
void InvolGenerator<Q>::processArgs(int, char **argv) {
    for (++argv; *argv; ++argv) {
        char *arg = *argv;
        if (arg[0] == '+') {
            hasOrigin = true;
        } else {
            currentInftySet.add((int) strtol(arg, nullptr, 10));
        }
    }
}

#endif //BSETS_INVOL_GENERATOR_H
