#ifndef BSETS_INVOL_INVOLGEOMETRY_H
#define BSETS_INVOL_INVOLGEOMETRY_H

#include "PointLineGeometry.hpp"
#include "Semi.hpp"
#include "Invol.hpp"

/**
 * Geometry of involution orbits. There are three types of points
 * <ul>
 * <li>The 'fat' points, numbered 0..F-1 (corresponding to orbits of size 2)
 * <li>The 'thin' points at infinity, numbered 0..Q (orbits of size 1)
 * <li>The 'origin' (also a 'thin' point).
 * </ul>
 * Three types of lines:
 * <ul>
 * <li>The 'fat' lines, numbered 0..F-1
 * <li>The 'thin' lines through the origin
 * <li>The 'thin' line at infinity
 * </ul>
 * Incidence is determined by a 'difference set' D
 * <ul>
 * <li>Fat line l contains a fat point f iff f-l ∈ D
 * <li>Fat line l contains point l % K at infinity
 * <li>Line l through the origin contains all points p = l mod K, and the origin (of course)
 * </ul>
 * Note that D contains exactly the points of line 0. Line l contains the points D+l (mod F).
 */

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/**
 * The point line geometry of fat and thin points and lines corresponding to
 * the orbits of an involution
 */
template<int Q>
class InvolGeometry : public PointLineGeometry<Invol<Q>::V, Invol<Q>::V> {

private:
    static constexpr int F = Invol<Q>::F;
    static constexpr int K = Invol<Q>::K;
    static constexpr int V = Invol<Q>::V;

    using PointLineGeometry<V,V>::setIncident;

public:
    InvolGeometry();

};

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////


template<int Q>
InvolGeometry<Q>::InvolGeometry() : PointLineGeometry<Invol<Q>::V, Invol<Q>::V>() {

    // for each FAT point
    for (int pt = 0; pt < F; ++pt) {
        // Q fat lines through pt
        for (int el: Semi<Q>::D) {
            setIncident(pt, (pt + 2 * F - el) % F);
        }
        // thin affine line through pt
        setIncident(pt, F + pt % K);
    }

    // for each point AT INFINITY
    for (int pt = 0; pt < K; ++pt) {
        // fat lines through pt
        for (int i = 0; i < (Q - 1) / 2; i++) {
            setIncident(F + pt, pt % K + i * K);
        }
        // line at infinity
        setIncident(F + pt, F + K);
        // thin line through pt
        setIncident(F + pt, F + pt);
    }

    // for THE ORIGIN

    // Q+1 thin lines through the origin
    for (int i = 0; i < K; ++i) {
        setIncident(F + K, F + i);
    }
}

#endif //BSETS_INVOL_INVOLGEOMETRY_H
