#ifndef BSETS_INVOL_WEIGHTS_HPP
#define BSETS_INVOL_WEIGHTS_HPP

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

#include <array>
#include "Invol.hpp"
#include "InvolSet.hpp"
#include "Semi.hpp"
#include "WeightsBase.hpp"

/**
 * Keeps track of the weights of the lines in the plane for a given point set
 */
template <int Q> 
class InvolWeights : public WeightsBase<Invol<Q>::V,Invol<Q>::F,InvolSet<Q>>{

private:
    static constexpr  int F = Invol<Q>::F;
    static constexpr  int K = Invol<Q>::K;
    static constexpr  int V = Invol<Q>::V;

    using WeightsBase<V,F,InvolSet<Q>>::set;
    using WeightsBase<V,F,InvolSet<Q>>::weights;
    using WeightsBase<V,F,InvolSet<Q>>::hashedWeight;

public:

    void add (int pt); // adding fat points
    void addInfty (int pt); // add a point at infinity, pt should be in 0,...,Q
    void addOrigin (); // add the origin

    void remove (int pt); //removing fat points
    void removeInfty (int pt);
    void removeOrigin ();

    bool isRedundantFat(int pt) const; //checks if a fat point is redundant; 0<=pt<F
    bool isRedundantInfty(int pt) const; //checks if a point at infinity is redundant; 0<=pt<K
    bool isRedundantOrigin() const; // checks if the origin is redundant

    // A point is superfluous iff all lines through it have weight ≥ 1
    bool isSuperfluous (int pt) const; //only called on fat points

    // A set is 'minimal' if there is no point for which all lines have weight > 1
    bool isMinimal() const ;

    // Compute and store the invariant values of all points of the set
    void computeInvariants () ;

    void printWeights() const;

private:
    // Compute invariant for the given FAT point. No need to compute invariants for the other points
    int computeInvariant (int pt) const;

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
void InvolWeights<Q>::add(int pt) {
    set.add(pt);
    // thin affine line through pt
    int m = pt % (K);
    weights[F + m] += 2; // fat point counts as 2
    // Q fat lines through pt
    for (int el: Semi<Q>::D) {
        int ln = (pt + 2 * F - el) % F;
        weights[ln]++;
    }

}

template<int Q>
void InvolWeights<Q>::addInfty(int pt) {// we assume that pt is in 0,...,Q
    set.addInfty(pt);
    weights[V - 1]++; //line at infinity
    weights[F + pt]++; //thin line through this point at infinity
    // Fat lines through pt
    int ln = pt % K; // ln is the fat line in 0,...,Q that goes through pt
    for (int i = 0; i < (Q - 1) / 2; ++i) {
        weights[ln + i * K]++;
    }
}

template<int Q>
void InvolWeights<Q>::addOrigin() {
    set.addOrigin();
    for (int i = 0; i < K; ++i) {
        weights[F + i]++; // raise weight of all affine thin lines by one.
    }
}

template<int Q>
void InvolWeights<Q>::remove(int pt) { //reverse what is done in add (fat point)
    // thin affine line through pt
    int m = pt % K;
    weights[F + m] -= 2; // fat point, counts as 2
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 2 * F - el) % F;
        weights[ln]--;
    }
    set.remove(pt);
}

template<int Q>
void InvolWeights<Q>::removeInfty(int pt) {// we assume that pt is in 0,...,Q
    weights[V - 1]--; //line at infinity
    weights[F + pt]--; //thin line through this point at infinity
    // Fat lines through pt
    int ln = pt % K; // ln is the fat line in 0,...,Q that goes through pt
    for (int i = 0; i < (Q - 1) / 2; ++i) {
        weights[ln + i * K]--;
    }
    set.removeInfty(pt);
}

template<int Q>
void InvolWeights<Q>::removeOrigin() {
    for (int i = 0; i < K; ++i) {
        weights[F + i]--; // lower weight of all affine thin lines by one.
    }
    set.removeOrigin();
}

template<int Q>
bool InvolWeights<Q>::isSuperfluous(int pt) const {
    // thin affine line through pt
    int m = pt % K;
    if (weights[F + m] == 0) { return false; }
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 2 * F - el) % F;
        if (weights[ln] == 0) { return false; }
    }
    return true;
}

template<int Q>
bool InvolWeights<Q>::isRedundantFat(int pt) const {
    // thin affine line through pt; well his weight is always at least 2, so no reason to check if weight <2
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 2 * F - el) % F;
        if (weights[ln] < 2) { return false; }
    }
    return true;
}

template<int Q>
bool InvolWeights<Q>::isRedundantInfty(int pt) const {
    if (weights[V - 1] < 2 || weights[F + pt] < 2) { return false; } // pt is the thin line through pt
    int ln = pt % K; // ln is the fat line in 0,...,Q that goes through pt
    for (int i = 0; i < (Q - 1) / 2; ++i) {
        if (weights[ln + i * K] < 2) {
            return false;
        }
    }
    return true;
}

template<int Q>
bool InvolWeights<Q>::isRedundantOrigin() const {
    for (int i = 0; i < K; ++i) {
        if (weights[F + i] < 2) {
            return false;
        }
    }
    return true;
}

template<int Q>
int InvolWeights<Q>::computeInvariant(int pt) const {
    int total = 0;
    // thin affine line through pt
    int m = pt % K;
    total += hashedWeight(F + m);
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 2 * F - el) % F;
        total += hashedWeight(ln);
    }
    return total;
}

template<int Q>
bool InvolWeights<Q>::isMinimal() const {

    // check whether any of the fat points is redundant
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt) && isRedundantFat(pt)) {
            return false;
        }
    }
    // check whether any of the points at infinity is redundant
    for (int pt = 0; pt < K; ++pt) {
        if (set.contains(F + pt) && isRedundantInfty(pt)) {
            return false;
        }
    }
    return !(set.contains(F + Q + 1) && isRedundantOrigin());
}

// TODO almost same as in Singer Weights
template<int Q>
void InvolWeights<Q>::computeInvariants() {
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt)) {
            this->invariants[pt] = computeInvariant(pt);
        }
    }
}

template<int Q>
void InvolWeights<Q>::printWeights() const {
    //printing weights
    std::cout << "WEIGHTS ";
    for (int j = 0; j <= K; ++j) {
        std::cout << j << ": ";
        for (int i = 0; i < V; ++i) {
            if (this->weights[i] == j) {
                std::cout << i << "-";
            }
        }
        std::cout << std::endl;
    }
}

#endif //BSETS_INVOL_WEIGHTS_HPP
