#ifndef BSETS_INVOL_WEIGHTS_WITH_SETH
#define BSETS_INVOL_WEIGHTS_WITH_SETH

#include <array>
#include "CheckResults.hpp"
#include "InvolSet.hpp"
#include "Semi.hpp"
#include "InvolWeights.hpp"

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/**
 * Keeps track of the weights of the lines in the plane wrt. the current point set
 */
template<int Q>
class InvolWeightsWithSet : public InvolWeights<Q> {

private:
    static constexpr int F = Invol<Q>::F;
    static constexpr int K = Invol<Q>::K;
    static constexpr int V = Invol<Q>::V;

    using InvolWeights<Q>::isRedundantFat;
    using InvolWeights<Q>::isRedundantInfty;
    using InvolWeights<Q>::isRedundantOrigin;
    using InvolWeights<Q>::computeInvariant;

    // Note about these arrays:
    // First F entries refer to the fat lines, next Q+1 entries to the thin lines, last entry is line at infinity
    std::array<int, V> invariants{};

    const PointSet<V> &set;

public:

    using InvolWeights<Q>::add;
    using InvolWeights<Q>::remove;

    explicit InvolWeightsWithSet(const PointSet<V> &set) : InvolWeights<Q>(), set(set) {

    }

    int invariantOf(int pt) const { return invariants[pt]; }

    // A set is 'minimal' if there is no point for which all lines have weight > 1
    bool isMinimal() const;

    // Compute and store the invariant values of all points of the set
    void computeInvariants();

    // Determine canonicity using the computed invariants
    // Call computeInvariants before calling this method
    CheckResult checkCanonical(int pt) const;

    // Compute the invariant for the complete set. A prior call to computeInvariants is necessary
    int computeSetInvariant() const;

    const int *invariantsData() const {
        return invariants.data();
    }

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
bool InvolWeightsWithSet<Q>::isMinimal() const {

    // check whether any of the fat points is redundant
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt) && isRedundantFat(pt)) {
            return false;
        }
    }
    // check whether any of the points at infinity is redundant
    for (int pt = 0; pt < K; ++pt) {
        if (set.contains(F + pt) && isRedundantInfty(pt)) {
            return false;
        }
    }
    return !(set.contains(F + Q + 1) && isRedundantOrigin());
}

template<int Q>
void InvolWeightsWithSet<Q>::computeInvariants() {
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt)) {
            invariants[pt] = computeInvariant(pt);
        }
    }
}

template<int Q>
CheckResult InvolWeightsWithSet<Q>::checkCanonical(int pt) const {
    // if the point has the highest unique invariant, then it is canonical
    // if there are no unique invariants, the point has to have the highest
    // invariant value to be able to be canonical.

    // Step 1. Check for highest unique invariant
    //////

    std::array<bool, F> removed{};
    // this array keeps track of which points should still be visited,
    // after the loop, only for the unique invariants the value remains false

    int nrOfUnique = 0;
    for (int i = 0; i < F; i++) {
        if (set.contains(i) && !removed[i]) {
            bool found = false;
            for (int j = i + 1; j < F; j++) {
                if (set.contains(j) && !removed[j] && invariants[i] == invariants[j]) {
                    found = true;
                    removed[j] = true;
                }
            }
            if (found) {
                removed[i] = true;
            } else {
                // i has a unique invariant value
                nrOfUnique++;
                if (invariants[i] > invariants[pt]) {
                    return NOT_CANONICAL; // there is at least one unique invariant value, but it is larger than for the requested point
                }
            }
        }
    }
    if (nrOfUnique != 0) {
        if (removed[pt]) {
            return NOT_CANONICAL;
        } else {
            return CANONICAL;
        }
    }

    // Step 2. Highest invariant value
    /////
    for (int i = 0; i < F; i++) {
        if (set.contains(i) && invariants[i] > invariants[pt]) {
            return NOT_CANONICAL;
        }
    }

    return UNKNOWN; // must defer to Nauty...
}

template<int Q>
int InvolWeightsWithSet<Q>::computeSetInvariant() const {
    int total = 0;
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt)) {
            total += invariants[pt] * invariants[pt];
        }
    }
    return total;
}

#endif //BSETS_INVOL_WEIGHTS_WITH_SETH
