#ifndef BSETS_LIB_GENERATORBASE_HPP
#define BSETS_LIB_GENERATORBASE_HPP

#include "CheckResults.hpp"
#include "ColourablePlane.hpp"
#include "Archive.hpp"

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/**
 * Common basis for the generators
 * @tparam WGHTS Class that stores the weights and the current set
 * @tparam V Total number of points in the plane
 * @tparam F Upper limit on the point index to be generated
 */
template<class WGHTS, int V, int F, int COUNTS_SIZE>
class GeneratorBase {

protected:

    WGHTS weights{};

    /**
     * Print the current set to stdout and return the number of points in
     * the projective plane which it represents;
     */
    virtual int shipSet() = 0;

    // add a point (which is not already in the current set)
    void add(int pt) {
        weights.add(pt);
    }

    // remove a point (which is already in the current set)
    void remove(int pt) {
        weights.remove(pt);
    }

private:
    ColourablePlane<V> colourablePlane;

    // check whether the currently generated set is complete
    bool isComplete() const {
        return weights.isBlocking();
    }

    // is this a valid subset of a possible result?
    bool isValid() const {
        return weights.isMinimal();
    }

    // is this a point that can be added to the current set so that it still remains valid?
    bool isValid(int pt) const {
        return !weights.isSuperfluous(pt);
    }

    // is the given point canonical in the set (wrt. isomorph freeness)
    CheckResult checkMaybeCanonical(int pt) {
        weights.computeInvariants();
        return weights.checkCanonical(pt);
    }

public:
    GeneratorBase(const PointLineGeometry<V, V> &geometry) :
            colourablePlane(ColourablePlane<V>(geometry)) {
    }

    void gen();

    // print statistics to stderr
    void printStatistics();

private:
    // nr of results generated
    long totalCount = 0;

    std::array<long, COUNTS_SIZE> counts = {0,}; // too large, but who cares?

protected:
    // Parameters to use when splitting process for parallelization
    int splitDepth = V + 1; // default = do not split
    int splitIndex = 0;
    int splitModulus = 0;
    int splitCounter = 0;

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<class WGHTS, int V, int F, int COUNT_SIZE>
void GeneratorBase<WGHTS, V, F, COUNT_SIZE>::gen() {
    const PointSet<V> &currentSet = weights.getSet();
        // bail out when not in the correct search branch (when parallelized)
    if (currentSet.getSize() == splitDepth) {
        if (splitCounter++ % splitModulus != splitIndex) {
            return;
        }
    }

    if (isComplete()) {
        totalCount++;
        counts[shipSet()]++;
    } else {
        Archive<ColourablePlane<V>, V> archive(colourablePlane);
        for (int pt = 0; pt < F; pt++) {
            if (!currentSet.contains(pt)) {
                if (isValid(pt)) {
                    add(pt);
                    if (isValid()) {
                        switch (checkMaybeCanonical(pt)) {
                            case CANONICAL: {
                                if (archive.add(currentSet, weights.computeSetInvariant())) {
                                    gen();
                                }
                                break;
                            }
                            case UNKNOWN: {
                                colourablePlane.colour(currentSet);
                                SparseGraphPtr canonicalForm
                                        = colourablePlane.isNautyCanonical(pt, weights.invariantsData());
                                if (canonicalForm) { // yes, it is special!
                                    if (archive.add(currentSet, weights.computeSetInvariant(),
                                                    std::move(canonicalForm))) {
                                        gen();
                                    }
                                }
                                break;
                            }
                            case NOT_CANONICAL: {
                                // do not recurse
                                break;
                            }
                        }
                    }
                    remove(pt);
                }
            }
        }
    }

}


template<class WGHTS, int V, int F, int COUNT_SIZE>
void GeneratorBase<WGHTS, V, F, COUNT_SIZE>::printStatistics() {
/*
    std::cerr << "Blocking sets of PG(2," << Q << ")" << std::endl;
    if (splitDepth <= V) {
        std::cerr << " parallellised - level " << splitDepth << ", index " << splitIndex << ", modulus " << splitModulus << std::endl;
    }
*/

    int start = 0;
    while (counts[start] == 0) {
        start++;
    }
    int stop = COUNT_SIZE-1;
    while (counts[stop] == 0) {
        stop--;
    }

    for (int i = start; i <= stop; i++) {
        std::cerr << "Size " << i << ": " << counts[i] << std::endl;
    }
}

#endif //BSETS_LIB_GENERATORBASE_HPP
