#ifndef BSETS_LINE_PROJECTIVITY_HPP
#define BSETS_LINE_PROJECTIVITY_HPP

#include <array>
#include "FFel.hpp"
#include "Point.hpp"
#include "../lib/PointSet.hpp"

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/**
 * Represents a projectivity of the projective line over the field of order Q.
 */
template<int Q>
class Projectivity {

public:
    using FF = FFel<Q>;
    using Matrix = std::array<std::array<FFel<Q>, 2>, 2>;
    const Matrix matrix;

    Point<Q> operator()(const Point<Q> &point) const;
    
    PointSet<Q+1> operator() (const PointSet<Q+1> &set) const;

    /**
     * Projectivity resulting from applying this projectivity
     * followed by the argument projectivity
     */
    Projectivity operator*(const Projectivity &other) const;

    Projectivity inv() const;

    bool isIdentity() const;

    /**
     * A projectivity that maps (0,1), (1,0) - i.e., the points with index 0,1 onto the given points.
     * <b>Important</b> Clients must ensure that the two points or not identical - this is not checked.
     */
    Projectivity<Q>(const Point<Q> &p0, const Point<Q> &p1) : Projectivity(
            {p1.x, p1.y, p0.x, p0.y}
    ) {}

    /**
     * A projectivity that maps (0,1), (1,0), (1,1) - i.e., the points with indices 0,1,2 = the standard base, onto the given points.
     * <b>Important</b> Clients must ensure that the three points are different - this is not checked.
     */
    Projectivity<Q>(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2)
            : Projectivity(matrixOfThree(p0, p1, p2)) {}

    /**
     * A projectivity that maps the given points onto the standard base.
     * <b>Important</b> Clients must ensure that the three points are different - this is not checked.
     */
    static Projectivity<Q> toBase(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2);

    /**
     * A projectivity with a given index. Can be used to loop over all projectivities. The number of projectivities is
     * (Q-1)Q(Q+1)
     */
    static Projectivity<Q> withIndex(int index);

    // TODO getIndex

private:

    Projectivity(const Matrix &matrix) : matrix(matrix) {
    }

    static Matrix matrixOfThree(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2);


};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
Point<Q> Projectivity<Q>::operator()(const Point<Q> &point) const {
    return Point{
            point.x * matrix[0][0] + point.y * matrix[1][0],
            point.x * matrix[0][1] + point.y * matrix[1][1]
    };
}

template<int Q>
Point<Q> Point<Q>::operator*(const Projectivity<Q> &projectivity) const {
    return projectivity(*this);
}

template<int Q>
Projectivity<Q> Projectivity<Q>::operator*(const Projectivity<Q> &other) const {
    Matrix result{};
    for (int i = 0; i < 2; ++i) {
        for (int j = 0; j < 2; ++j) {
            result[i][j] =
                    matrix[i][0] * other.matrix[0][j] +
                    matrix[i][1] * other.matrix[1][j];
        }
    }
    return Projectivity<Q>(result);
}

template<int Q>
bool Projectivity<Q>::isIdentity() const {
    return matrix[0][1].isZero() && matrix[1][0].isZero() && matrix[0][0] == matrix[1][1];
}

template<int Q>
Projectivity<Q> Projectivity<Q>::inv() const {
    return Projectivity({
                                matrix[1][1], -matrix[0][1],
                                -matrix[1][0], matrix[0][0],
                        });
}

template<int Q>
std::array<std::array<FFel<Q>, 2>, 2>
Projectivity<Q>::matrixOfThree(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2) {
    Projectivity result{p0, p1};
    Point<Q> d = p2 * result.inv();
    return {result.matrix[0][0] * d.x, result.matrix[0][1] * d.x,
            result.matrix[1][0] * d.y, result.matrix[1][1] * d.y,
    };
}

template<int Q>
Projectivity<Q>
Projectivity<Q>::toBase(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2) {
    Projectivity result = Projectivity(p0, p1).inv();
    Point<Q> d = p2 * result;
    return Projectivity({result.matrix[0][0] / d.x, result.matrix[0][1] / d.y,
                         result.matrix[1][0] / d.x, result.matrix[1][1] / d.y,
                        });
}

template<int Q>
Projectivity<Q> Projectivity<Q>::withIndex(int index) {
    if (index < Q * Q * (Q - 1)) {
        // [ 1  a   ]
        // [ b ab+c ] c ≠ 0
        FF c = FF::get(1 + index % (Q - 1));
        index /= Q - 1;
        FF b = FF::get(index % Q);
        FF a = FF::get(index / Q);
        return Projectivity({FF(1), a, b , a * b + c,});
    } else {
        // [ 0 1 ]
        // [ a b ]  a ≠ 0
        index -= Q * Q * (Q - 1);
        FF b = FF::get(index % Q);
        FF a = FF::get(1 + index / Q);
        return Projectivity({FF(0), FF(1), a, b,});
    }
}

template<int Q>
PointSet<Q+1> Projectivity<Q>::operator()(const PointSet<Q+1> &set) const {
    PointSet<Q+1> result{};
    int size = set.getSize();
    for (int pt=0, cnt=0; pt <= Q && cnt < size; ++pt) {
        if (set.contains(pt)) {
            cnt ++;
            result.add(operator()(Point<Q>::get(pt)).index());
        }
    }
    return result;
}


#endif //BSETS_LINE_PROJECTIVITY_HPP
