#ifndef BSETS_PLANE_CANONICALFORM_HPP
#define BSETS_PLANE_CANONICALFORM_HPP

#include <array>
#include <algorithm>
#include "PointSet.hpp"
#include "WeightsBase.hpp"
#include "Projectivity.hpp"

/*
 * Provides methods to produce a canonical form of a set (w.r.t. the projective group of the plane)
 */

/**
 * Point-invariant pair.
 */
template<int Q>
struct PtInvar {
    const Point<Q> *pt; // TODO make these references
    int invariant;

//    PtInvar(const Point<Q> *pt, int invariant) : pt (pt), invariant(invariant) {
//
//    }

    bool operator<(const PtInvar &other) const {
        return invariant < other.invariant;
    }
};

template<int Q>
class CanonicalForm {

public:
    static constexpr int V = Q * Q + Q + 1;

private:
    std::array<PtInvar<Q>, V> points;
    const PointSet<V> &set;
    PointSet<V> canonicalSet;
    int groupSize{0};

    // auxiliary functions used in constructor

    void sortByMultiplicity();

protected:
    virtual void processImage (const PointSet<V> &image);

public:
    /**
     * Should be called with a weights base for which the invariants for *all* points have
     * been precomputed.
     */
    CanonicalForm(const WeightsBase<V> &base);

    /**
     * Computes the canonical form of the current set.
     */
    void compute();

    /**
     * The computed canonical set
     */
    const PointSet<V> &getCanonicalSet() const {
        return canonicalSet;
    }

    /**
     * The computed group size
     */
    int getGroupSize() const {
        return groupSize;
    }

    // mainly used for testing
    void printInvariants() const;
};

// for debugging purposes
template<int Q>
void CanonicalForm<Q>::printInvariants() const {
    for (int pt = 0; pt < V; ++pt) {
        std::cout << " " << points[pt].invariant;
    }
    std::cout << std::endl;
}

template<int Q>
CanonicalForm<Q>::CanonicalForm(const WeightsBase<V> &base)  : set(base.getSet()) {
    for (int pt = 0; pt < V; ++pt) {
        points[pt].pt = &Point<Q>::get(pt);
        points[pt].invariant = base.invariantsData()[pt];
    }
    // sort according to weight
    std::sort(points.begin(), points.end());
    // now sort according to multiplicities
    sortByMultiplicity();
}

template<int Q>
void CanonicalForm<Q>::sortByMultiplicity() {
    // starts with an array points sorted according to invariant value

    // determine run lengths
    std::array<int, V> runLengths; // for each point, how many times its invariant occurs
    std::array<int, V + 1> freq{};  // for each run length how many times it occurs
    // multiplied by the run length itself
    int start{0};
    for (int i = 1; i < V; i++) {
        if (points[i].invariant != points[start].invariant) {
            // end of run
            int run = i - start;
            for (int j = start; j < i; j++) {
                runLengths[j] = run;
            }
            freq[run] += run;
            start = i;
        }
    }
    // last run
    int run = V - start;
    for (int j = start; j < V; j++) {
        runLengths[j] = run;
    }
    freq[run] += run;

    // accumulate frequencies
    for (int i = 1; i <= V; i++) {
        freq[i] += freq[i - 1];
    }
    // at this point freq[r] indicates the output position -1
    // of the last point with this runlength

    // relocate
    std::array<PtInvar<Q>, V> newPoints{};
    for (int j = V - 1; j >= 0; --j) {
        int r = runLengths[j];
        int pos = freq[r] - 1;
        freq[r] = pos;
        newPoints[pos] = points[j];
    }
    points = newPoints;
}

template<int Q>
void CanonicalForm<Q>::compute() {
    canonicalSet = PointSet<V>(); // initial best fit
    groupSize = 0;

    int inv0 = points[0].invariant;
    // pt0 can be any of the points with invariant inv0
    for (int i0 = 0; points[i0].invariant == inv0 && i0 < V; ++i0) {
        const Point<Q> &point0 = *points[i0].pt;
        // starting value for pt1 is the first point different from pt0
        int i1 = i0 == 0 ? 1 : 0;
        int inv1 = points[i1].invariant;
        int sg1 = inv0 == inv1 ? 0 : i1; // start of the group with invariant inv1
        // pt1 can be any of the points with invariant inv1, different from pt0
        for (; points[i1].invariant == inv1 && i1 < V; ++i1) {
            if (i1 != i0) {
                // starting value for pt2 is the first point 'not later' than pt1 which does not lie
                // on the line pt0/pt1
                const Point<Q> &point1 = *points[i1].pt;
                const Point<Q> line01 = point0.lineThrough(point1);
                int i2 = sg1;
                while (i2 == i0 || i2 == i1 || points[i2].pt->liesOn(line01)) {
                    ++i2;
                }
                int inv2 = points[i2].invariant;
                int sg2 = inv1 == inv2 ? sg1 : i2; // start of the group with invariant inv2
                // pt2 can be any of the points with invariant inv2, not on the line pt0/pt1
                for (; points[i2].invariant == inv2 && i2 < V; ++i2) {
                    if (!points[i2].pt->liesOn(line01)) {
                        const Point<Q> &point2 = *points[i2].pt;
                        const Point<Q> line02 = point0.lineThrough(point2);
                        const Point<Q> line12 = point1.lineThrough(point2);
                        // point pt3 should not lie on the triangle formed by the three other points
                        int i3 = sg2;
                        while (i3 == i0 || i3 == i1 || i3 == i2
                               || points[i3].pt->liesOn(line01)
                               || points[i3].pt->liesOn(line02)
                               || points[i3].pt->liesOn(line12)) {
                            ++i3;
                        }
                        int inv3 = points[i3].invariant;
                        for (; points[i3].invariant == inv3 && i3 < V; ++i3) {
                            const Point<Q> &point3 = *points[i3].pt;
                            if (i3 != i0 && i3 != i1 && i3 != i2
                                && !point3.liesOn(line01)
                                && !point3.liesOn(line02)
                                && !point3.liesOn(line12)) {
//                                printf("%d %d %d %d\n", i0, i1, i2, i3);
                                Projectivity<Q> proj = Projectivity<Q>::toBase(point0, point1, point2, point3);
                                PointSet<V> image = proj(set);
                                processImage(image);
                            }
                        }
                    }
                }
            }
        }
    }

}

template<int Q>
void CanonicalForm<Q>::processImage(const PointSet<V> &image) {
    int c = canonicalSet.cmp(image);
    if (c == 0) {
        // still canonical
        groupSize++;
    } else if (c < 0) {
        // new canonical form
        groupSize = 1;
        canonicalSet = image;
    }
}


#endif //BSETS_PLANE_CANONICALFORM_HPP
