#ifndef BSETS_PLANE_PROJECTIVITY_HPP
#define BSETS_PLANE_PROJECTIVITY_HPP

#include <array>
#include "FFel.hpp"
#include "Point.hpp"
#include "PointSet.hpp"
#include "Partition.hpp"

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

/**
 * Represents a projectivity of the projective plane
 * over the field of order Q.
 */
template<int Q>
class Projectivity {

public:
    using FF = FFel<Q>;
    using Matrix = std::array<std::array<FFel<Q>, 3>, 3>;
    const Matrix matrix;

    Point<Q> operator()(const Point<Q> &point) const;

    int operator()(int point) const {
        return operator()(Point<Q>::get(point)).index();
    }

    PointSet<Q*Q+Q+1> operator() (const PointSet<Q*Q+Q+1> &point) const;

    /**
     * Projectivity resulting from applying this projectivity
     * followed by the argument projectivity
     */
    Projectivity operator*(const Projectivity &other) const;

    Projectivity inv() const;

    bool isIdentity() const;

    /**
     * A projectivity that maps (0,0,1), (0,1,0), (1,0,0) - i.e., the points with index 0,1,Q+1 onto the given points.
     * <b>Important</b> Clients must ensure that the three points are not collinear - this is not checked.
     */
    Projectivity<Q>(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2) : Projectivity(
            {p2.x, p2.y, p2.z, p1.x, p1.y, p1.z, p0.x, p0.y, p0.z}
    ) {}

    /**
     * A projectivity that maps (0,0,1), (0,1,0), (1,0,0), (1,1,1) - i.e., the standard base, onto the given points.
     * <b>Important</b> Clients must ensure that of the four points no three are collinear - this is not checked.
     */
    Projectivity<Q>(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2, const Point<Q> &p3)
            : Projectivity(matrixOfFour(p0, p1, p2, p3)) {}

    /**
     * A projectivity that maps the given points onto the standard base.
     * <b>Important</b> Clients must ensure that of the four points no three are collinear - this is not checked.
     */
    static Projectivity<Q> toBase(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2, const Point<Q> &p3);

    /**
     * Orbit partition corresponding to this projectivity
     */
    Partition<Q*Q+Q+1> orbitPartition() const;


    Projectivity(const Matrix &matrix) : matrix(matrix) {
    }

private:

    static Matrix matrixOfFour(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2, const Point<Q> &p3);


};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
Point<Q> Projectivity<Q>::operator()(const Point<Q> &point) const {
    return Point{
            point.x * matrix[0][0] + point.y * matrix[1][0] + point.z * matrix[2][0],
            point.x * matrix[0][1] + point.y * matrix[1][1] + point.z * matrix[2][1],
            point.x * matrix[0][2] + point.y * matrix[1][2] + point.z * matrix[2][2]
    };
}

template<int Q>
Point<Q> Point<Q>::operator*(const Projectivity<Q> &projectivity) const {
    return projectivity(*this);
}

template<int Q>
Projectivity<Q> Projectivity<Q>::operator*(const Projectivity<Q> &other) const {
    Matrix result{};
    for (int i = 0; i < 3; ++i) {
        for (int j = 0; j < 3; ++j) {
            result[i][j] =
                    matrix[i][0] * other.matrix[0][j] +
                    matrix[i][1] * other.matrix[1][j] +
                    matrix[i][2] * other.matrix[2][j];
        }
    }
    return Projectivity<Q>(result);
}

template<int Q>
bool Projectivity<Q>::isIdentity() const {
    for (int i = 0; i < 3; ++i) {
        for (int j = i + 1; j < 3; ++j) {
            if (!matrix[i][j].isZero() || !matrix[j][i].isZero()) {
                return false;
            }
        }
    }
    return matrix[0][0] == matrix[1][1] && matrix[1][1] == matrix[2][2];
}

template<int Q>
Projectivity<Q> Projectivity<Q>::inv() const {
    return Projectivity{
            {
                    matrix[1][1] * matrix[2][2] - matrix[1][2] * matrix[2][1],
                    matrix[2][1] * matrix[0][2] - matrix[2][2] * matrix[0][1],
                    matrix[0][1] * matrix[1][2] - matrix[0][2] * matrix[1][1],
                    matrix[1][2] * matrix[2][0] - matrix[1][0] * matrix[2][2],
                    matrix[2][2] * matrix[0][0] - matrix[2][0] * matrix[0][2],
                    matrix[0][2] * matrix[1][0] - matrix[0][0] * matrix[1][2],
                    matrix[1][0] * matrix[2][1] - matrix[1][1] * matrix[2][0],
                    matrix[2][0] * matrix[0][1] - matrix[2][1] * matrix[0][0],
                    matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0],
            },
    };
}

template<int Q>
std::array<std::array<FFel<Q>, 3>, 3>
Projectivity<Q>::matrixOfFour(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2, const Point<Q> &p3) {
    Projectivity result{p0, p1, p2};
    Point<Q> d = p3 * result.inv();
    return {result.matrix[0][0] * d.x, result.matrix[0][1] * d.x, result.matrix[0][2] * d.x,
            result.matrix[1][0] * d.y, result.matrix[1][1] * d.y, result.matrix[1][2] * d.y,
            result.matrix[2][0] * d.z, result.matrix[2][1] * d.z, result.matrix[2][2] * d.z,
    };
}

template<int Q>
Projectivity<Q>
Projectivity<Q>::toBase(const Point<Q> &p0, const Point<Q> &p1, const Point<Q> &p2, const Point<Q> &p3) {
    Projectivity result = Projectivity(p0, p1, p2).inv();
    Point<Q> d = p3 * result;
    return Projectivity({result.matrix[0][0] / d.x, result.matrix[0][1] / d.y, result.matrix[0][2] / d.z,
                         result.matrix[1][0] / d.x, result.matrix[1][1] / d.y, result.matrix[1][2] / d.z,
                         result.matrix[2][0] / d.x, result.matrix[2][1] / d.y, result.matrix[2][2] / d.z,
                        });
}
template<int Q>
PointSet<Q*Q+Q+1> Projectivity<Q>::operator()(const PointSet<Q*Q+Q+1> &set) const {
    PointSet<Q*Q+Q+1> result{};
    int size = set.getSize();
    for (int pt=0, cnt=0; pt < Q*Q+Q+1 && cnt < size; ++pt) {
        if (set.contains(pt)) {
            cnt ++;
            result.add(operator()(Point<Q>::get(pt)).index());
        }
    }
    return result;
}

template<int Q>
Partition<Q * Q + Q + 1> Projectivity<Q>::orbitPartition() const {
    Partition<Q*Q+Q+1> partition{};
    for (int el = 0; el < Q*Q+Q+1; el++) {
        if (partition.partOf(el) < 0) {
            int part = partition.addPart();
            partition.add(part, el);
            int image = (*this)(el);
            while (image != el) {
                partition.add(part, image);
                image = (*this)(image);
            }
        }
    }
    return partition;
}

#endif //BSETS_PLANE_PROJECTIVITY_HPP
