#ifndef BSETS_PLANE_WEIGHTS_HPP
#define BSETS_PLANE_WEIGHTS_HPP

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

#include <array>
#include "WeightsBase.hpp"
#include "CheckResults.hpp"
#include "ProjectivePlane.hpp"

/**
 * Keeps track of the weights of the lines in the plane for a given point set
 */
template<int Q>
class Weights : public WeightsBase<Q*Q+Q+1, Q*Q+Q+1> {

private:
    static constexpr int V = Q*Q+Q+1;

    std::array<std::array<int,Q+1>,V> members; // points on a line = lines on a point

    using WeightsBase<V>::set;
    using WeightsBase<V>::weights;
    using WeightsBase<V>::hashedWeight;

public:

    Weights(const ProjectivePlane<Q> &pp);

    void add(int pt);

    void remove(int pt);

    // A point is redundant when all lines through that point have weight > 1
    bool isRedundant(int pt) const;

    // A point is superfluous iff all lines through it have weight ≥ 1
    bool isSuperfluous(int pt) const;

    // A set is 'minimal' if there is no point for which all lines have weight > 1
    bool isMinimal() const;

    // Compute and store the invariant values of all points of the set
    void computeInvariants();

    // Compute and store the invariant values of all points of the plane - not only the current set
    void computeAllInvariants();

    void printWeights() const;

protected:
    // Compute invariant for the point.
    int computeInvariant(int pt) const;

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
Weights<Q>::Weights(const ProjectivePlane<Q> &pp) {

    for (int pt=0; pt < V; ++pt) {
        int pos = 0;
        for (int ln = 0; ln < V; ++ln) {
            if (pp.isIncident(pt,ln)) {
                members[pt][pos++] = ln;
            }
        }
    }
}

template<int Q>
void Weights<Q>::add(int pt) {
    set.add(pt);
    for (int ln: members[pt]) {
        weights[ln]++;
    }
}

template<int Q>
void Weights<Q>::remove(int pt) { // reverse what is done in add
    for (int ln: members[pt]) {
        weights[ln]--;
    }
    set.remove(pt);
}

template<int Q>
bool Weights<Q>::isSuperfluous(int pt) const {
    for (int ln: members[pt]) {
        if (weights[ln] == 0) {
            return false;
        }
    }
    return true;
}

template<int Q>
bool Weights<Q>::isRedundant(int pt) const {
    for (int ln: members[pt]) {
        if (weights[ln] < 2) {
            return false;
        }
    }
    return true;
}

template<int Q>
bool Weights<Q>::isMinimal() const {
    // check whether any point is redundant
    for (int pt = 0; pt < V; pt++) {
        if (set.contains(pt) && isRedundant(pt)) {
            return false;
        }
    }
    return true;
}

template<int Q>
int Weights<Q>::computeInvariant(int pt) const {
    int total = 0;
    for (int el: members[pt]) {
        total += hashedWeight(el);
    }
    return total;
}

template<int Q>
void Weights<Q>::computeInvariants() {
    for (int pt = 0; pt < V; pt++) {
        if (set.contains(pt)) {
            this->invariants[pt] = computeInvariant(pt);
        }
    }
}

template<int Q>
void Weights<Q>::computeAllInvariants() {
    for (int pt = 0; pt < V; ++pt) {
        this->invariants[pt] = computeInvariant(pt);
    }
}

template<int Q>
void Weights<Q>::printWeights() const {
    std::cout << "WEIGHTS ";
    for (int w = 0; w <= Q + 1; ++w) {
        std::cout << w << ": ";
        for (int i = 0; i < V; ++i) {
            if (this->weights[i] == w) {
                std::cout << i << "-";
            }
        }
        std::cout << std::endl;
    }
}

#endif //BSETS_PLANE_WEIGHTS_HPP
