#ifndef BSETS_SEMI_SEMI_HPP
#define BSETS_SEMI_SEMI_HPP

#include <array>
#include "FFel.hpp"
#include "../plane/Point.hpp"

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

enum SemiPointType {
    ORIGIN, AT_INFTY, SEMI_AFFINE
};

template<int Q>
class Semi {
public:
    static const std::array<int, Q> D;

    using FF = FFel<Q>;

    /**
     * Return the point with given index, as a point of the plane
     */
    static Point<Q> point(int index) {
        static std::array<Point<Q>, Q * Q - 1> points = initPoints();
        return points[index];
    }

    /**
     * Return the point with given index, as a point of the line at infinity x=0
     */
    static Point<Q> atInfty(int index) {
        static std::array<Point<Q>, Q + 1> points = initAtInfty();
        return points[index];
    }

    /**
     * The origin, i.e., the point with coordinates (1,0,0);
     */
    static const Point<Q> origin;

    /**
     * Convert a point of the plane to an index in the semi-affine plane
     */
    static std::pair<SemiPointType, int> index(int pointIndex) {
        static std::array<std::pair<SemiPointType, int>, Q * Q + Q + 1> results = initIndices();
        return results[pointIndex];
    }

private:

    // image of (0,1) by defining cycle is (a,b)
    static const FF a;
    static const FF b;

    static std::array<int, Q> initD(FF y0);

    static bool checkDiffset(std::array<int, Q> &diffset);

    static std::array<Point<Q>, Q * Q - 1> initPoints();

    static std::array<Point<Q>, Q + 1> initAtInfty();

    static std::array<std::pair<SemiPointType, int>, Q * Q + Q + 1> initIndices();

};

////////////////////////////////////////
// DEFINITIONS - GENERAL
////////////////////////////////////////

// TODO: initD can always use 1 as a parameter!
template<int Q>
std::array<int, Q> Semi<Q>::initD(FF y0) {
    std::array<int, Q> result{};
    int pos{0};
    FF x = 1;
    FF y = 0;
    for (int el = 0; pos < Q; el++) {
        if (y == y0) {
            result[pos] = el;
            ++pos;
        }
        FF newx = y;
        y = a * x + b * y;
        x = newx;
    }
    if (checkDiffset(result)) {
        return result;
    } else {
        throw std::logic_error("Difference set does not satisfy required properties");
    }
}

template<int Q>
bool Semi<Q>::checkDiffset(std::array<int, Q> &diffset) {
    constexpr int V = Q * Q - 1;
    constexpr int K = Q + 1;

    std::array<bool, V> present{};
    for (const int i: diffset) {
        for (const int j: diffset) {
            present[(V + i - j) % V] = true;
        }
    }
    for (int i = 1; i < V; i++) {
        if (present[i] == (i % K == 0)) {
            return false;
        }
    }
    for (const int el: diffset) {
        if (el % K == 0) {
            return false;
        }
    }
    return true;
}

template<int Q>
std::array<Point<Q>, Q * Q - 1> Semi<Q>::initPoints() {
    std::array<Point<Q>, Q * Q - 1> result{};
    FF x(1);
    FF y(0);
    for (int i = 0; i < Q * Q - 1; ++i) {
        result[i] = Point<Q>(FF(1), x, y);
        FF newx = y;
        y = a * x + b * y;
        x = newx;
    }
    return result;
}

template<int Q>
std::array<Point<Q>, Q + 1> Semi<Q>::initAtInfty() {
    std::array<Point<Q>, Q + 1> result{};
    FF x(1);
    FF y(0);
    for (int i = 0; i <= Q; ++i) {
        result[i] = x.isZero() ? Point<Q>(FF(0), FF(0), FF(1)) : Point<Q>(FF(0), FF(1), y / x);
        FF newx = y;
        y = a * x + b * y;
        x = newx;
    }
    return result;
}

template<int Q>
const Point<Q> Semi<Q>::origin{FF(1), FF(0), FF(0)};

template<int Q>
std::array<std::pair<SemiPointType, int>, Q * Q + Q + 1> Semi<Q>::initIndices() {
    std::array<std::pair<SemiPointType, int>, Q * Q + Q + 1> result{};
    for (int pt = 0; pt < Q * Q - 1; ++pt) {
        result[point(pt).index()] = std::pair(SEMI_AFFINE, pt);
    }
    for (int pt=0; pt < Q+1; ++pt) {
        result[atInfty(pt).index()] = std::pair(AT_INFTY, pt);
    }
    result[origin.index()] = std::pair(ORIGIN, 0);
    return result;
}

////////////////////////////////////////
// DEFINITIONS - SPECIALIZATIONS
////////////////////////////////////////

// Generated by matrix  (0  1\\1 2), point 0: (1,0), line 0: Y=1   (field = Z/3Z)
// 1, 6, 7 mod 8
template<> const FFel<3> Semi<3>::a(1);
template<> const FFel<3> Semi<3>::b(2);
template<> const std::array<int, 3>  Semi<3>::D = initD(FF(1));

// Generated by matrix  (0  1\\3  4), point 0: (1,0), line 0: Y=2   (field = Z/5Z)
// 2, 3, 5, 13, 22 mod 24
template<> const FFel<5> Semi<5>::a(3);
template<> const FFel<5> Semi<5>::b(4);
template<> const std::array<int, 5>  Semi<5>::D = initD(FF(2));

// Generated by matrix  (0  1\\4  6), point 0: (1,0), line 0: Y=2   (field = Z/7Z)
// 10, 15, 33, 37, 43, 44, 46 mod 48
template<> const FFel<7> Semi<7>::a(4);
template<> const FFel<7> Semi<7>::b(6);
template<> const std::array<int, 7>  Semi<7>::D = initD(FF(2));

// Generated by matrix (0 1\\α³ 1), point 0: (1,0), line 0: Y=1   (field = GF(8))
// 7, 12, 23, 33, 55, 56, 58, 62 mod 63
template<> const FFel<8> Semi<8>::a = FFel<8>::all()[4];
template<> const FFel<8> Semi<8>::b(1);
template<> const std::array<int, 8> Semi<8>::D = initD(FF(1));

// Generated by matrix  (0  1\\i+1  1), point 0: (1,0), line 0: Y=1   (field = GF(9))
// 3, 6, 8, 14, 31, 32, 45, 67, 79 mod 80
template<> const FFel<9> Semi<9>::a = FFel<9>::all()[2];
template<> const FFel<9> Semi<9>::b(1);
template<> const std::array<int, 9> Semi<9>::D = initD(FF(1));

// Generated by matrix  (0  1\\4  10), point 0: (1,0), line 0: Y=3   (field = Z/11Z)
// 38, 44, 45, 47, 64, 75, 79, 89, 97, 102 mod 120
template<> const FFel<11> Semi<11>::a(4);
template<> const FFel<11> Semi<11>::b(10);
template<> const std::array<int, 11> Semi<11>::D = initD(FF(3));

// generate by matrix (0 1 \\ 6 2), point 0: (1,0), line 0: Y=1 (field = GF(13))
// 6, 26, 37, 45, 61, 71, 86, 108, 144, 161, 162, 165, 167 mod 168
template<> const FFel<13> Semi<13>::a(6);
template<> const FFel<13> Semi<13>::b(2);
template<> const std::array<int, 13> Semi<13>::D = initD(FF(1));



#endif //BSETS_SEMI_SEMI_HPP
