#ifndef BSETSNONTRIV_THREEGENERATOR_HPP
#define BSETSNONTRIV_THREEGENERATOR_HPP

#include <iostream>
#include <fstream>

#include "ThreeSet.hpp"
#include "ThreeWeights.hpp"
#include "Archive.hpp"
#include "ColourablePlane.hpp"
#include "ThreeGeometry.hpp"
#include "GeneratorBase.hpp"

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

template<int Q>
class ThreeGenerator :
        public GeneratorBase<ThreeWeights<Q>, Three<Q>::V, Three<Q>::F, Q*Q+Q+1> {

private:
    static constexpr int F = Three<Q>::F;
    static constexpr int V = Three<Q>::V;
    static constexpr int K = Three<Q>::K;

public:

    ThreeGenerator() :
            GeneratorBase<ThreeWeights<Q>, V, F, Q*Q+Q+1>(ThreeGeometry<Q>()) {

    }

    void generate();

private:
    using GeneratorBase<ThreeWeights<Q>, V, F, Q*Q+Q+1>::weights;
    using GeneratorBase<ThreeWeights<Q>, V, F, Q*Q+Q+1>::gen;
    using GeneratorBase<ThreeWeights<Q>, V, F, Q*Q+Q+1>::shipSet;

    // the current set of points at infinity
    PointSet<K> currentInftySet{};

    // whether the origin belongs to the set
    bool hasOrigin{false};

    void addOrigin(); // only used for a starting configuration
    void addInfty(int pt); // only used for a starting configuration
    void removeInfty(int pt);
    void removeOrigin();

    virtual int shipSet();

public:

    // process arguments - they contain the points at infinity to be included, and/or the origin
    void processArgs(int arg, char **argv);

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
void ThreeGenerator<Q>::generate() {
    // starting configuration was added in processArgs
    if (hasOrigin) {
        addOrigin();
    }
    for (int pt = 0; pt < K; ++pt) {
        if (currentInftySet.contains(pt)) {
            addInfty(pt);
        }
    }

    // recursively generate all fat points
    gen();

    // undo starting configuration
    for (int pt = K-1; pt >= 0; --pt) {
        if (currentInftySet.contains(pt)) {
            removeInfty(pt);
        }
    }
    if (hasOrigin) {
        removeOrigin();
    }
}


template<int Q>
int ThreeGenerator<Q>::shipSet() {
    int size{0};
    // print to stdout
    if (hasOrigin) {
        printf("+0");
        size++;
    }
    for (int pt = 0; pt < K; ++pt) {
        if (currentInftySet.contains(pt)) {
            printf("-%d-%d-%d", pt, pt+K, pt+2*K);
            size += 3;
        }
    }
    for (int pt = 0; pt < F; pt++) { // fat affine points
        if (weights.getSet().contains(pt)) {
            printf(":%d:%d:%d", pt, pt+F, pt+2*F);
            size += 3;
        }
    }
    printf("\n");
    return size;
}

template<int Q>
void ThreeGenerator<Q>::addOrigin() {
    weights.addOrigin();
}

template<int Q>
void ThreeGenerator<Q>::addInfty(int pt) {
    weights.addInfty(pt);
}

template<int Q>
void ThreeGenerator<Q>::removeOrigin() {
    weights.removeOrigin();
}

template<int Q>
void ThreeGenerator<Q>::removeInfty(int pt) {
    weights.removeInfty(pt);
}

template<int Q>
void ThreeGenerator<Q>::processArgs(int, char **argv) {
    for (++argv; *argv; ++argv) {
        char *arg = *argv;
        if (arg[0] == '+') {
            hasOrigin = true;
        } else {
            currentInftySet.add((int) strtol(arg, nullptr, 10));
        }
    }
}

#endif //BSETSNONTRIV_THREEGENERATOR_HPP
