#ifndef BSETSNONTRIV_THREEGEOMETRY_HPP
#define BSETSNONTRIV_THREEGEOMETRY_HPP

#include "PointLineGeometry.hpp"
#include "Semi.hpp"
#include "Three.hpp"

/**
 * Geometry of orbits under an automorphism of order three working on PG(2,q); q = 2 (mod 3). There are three types of points
 * <ul>
 * <li>The 'fat' affine points, numbered 0..F-1 (corresponding to orbits of size 3)
 * <li>The fat points at infinity, numbered 0..K-1 (orbits of size 3 on the line at infinity)
 * <li>The 'origin' (a 'thin' point).
 * </ul>
 * Three types of lines:
 * <ul>
 * <li>The 'fat' affine lines, numbered 0..F-1
 * <li>The fat lines through the origin, numbered 0..K-1
 * <li>The 'thin' line at infinity
 * </ul>
 * Incidence is determined by a 'difference set' D
 * <ul>
 * <li>Fat line l contains a fat point f iff f-l ∈ D
 * <li>Fat line l contains point l % K at infinity
 * <li>Line l through the origin contains all points p = l mod K, and the origin (of course)
 * </ul>
 * Note that D contains exactly the points of line 0. Line l contains the points D+l (mod F).
 */

////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

template<int Q>
class ThreeGeometry : public PointLineGeometry<Three<Q>::V, Three<Q>::V> {

private:
    static constexpr int F = Three<Q>::F;
    static constexpr int K = Three<Q>::K;
    static constexpr int V = Three<Q>::V;

    using PointLineGeometry<V,V>::setIncident;

public:
    ThreeGeometry();
};

////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
ThreeGeometry<Q>::ThreeGeometry() : PointLineGeometry<Three<Q>::V, Three<Q>::V>() {
    // for each FAT point
    for (int pt = 0; pt < F; ++pt) {
        // Q fat lines through pt
        for (int el: Semi<Q>::D) {
            setIncident(pt, (pt + 3 * F - el) % F);
        }
        // fat line through origin
        setIncident(pt, F + pt % K);
    }

    // for each point AT INFINITY
    for (int pt = 0; pt < K; ++pt) {
        // fat lines through pt
        for (int i = 0; i < Q-1; i++) {
            setIncident(F + pt, pt + i * K);
        }
        // line at infinity
        setIncident(F + pt, V-1);
        // line through origin
        setIncident(F + pt, F + pt);
    }

    // for THE ORIGIN
    // K fat lines through the origin
    for (int i = 0; i < K; ++i) {
        setIncident(V-1, F + i);
    }
}

#endif //BSETSNONTRIV_THREEGEOMETRY_HPP
