#ifndef BSETSNONTRIV_THREEWEIGHTS_HPP
#define BSETSNONTRIV_THREEWEIGHTS_HPP

/////////////////////////////////////////
// DECLARATIONS
////////////////////////////////////////

#include <array>
#include "Three.hpp"
#include "ThreeSet.hpp"
#include "Semi.hpp"
#include "WeightsBase.hpp"

/**
 * Keeps track of the weights of the lines in the plane for a given point set
 */
template <int Q>
class ThreeWeights : public WeightsBase<Three<Q>::V,Three<Q>::F,ThreeSet<Q>>{

private:
    static constexpr  int F = Three<Q>::F;
    static constexpr  int K = Three<Q>::K;
    static constexpr  int V = Three<Q>::V;

    using WeightsBase<V,F,ThreeSet<Q>>::set;
    using WeightsBase<V,F,ThreeSet<Q>>::weights;
    using WeightsBase<V,F,ThreeSet<Q>>::hashedWeight;

public:

    void add (int pt); // adding fat points
    void addInfty (int pt); // add a point at infinity, pt should be in 0,...,Q
    void addOrigin (); // add the origin

    void remove (int pt); //removing fat points
    void removeInfty (int pt);
    void removeOrigin ();

    bool isRedundantFat(int pt) const; //checks if a fat point is redundant; 0<=pt<F
    bool isRedundantInfty(int pt) const; //checks if a point at infinity is redundant; 0<=pt<K
    bool isRedundantOrigin() const; // checks if the origin is redundant

    // A point is superfluous iff all lines through it have weight ≥ 1
    bool isSuperfluous (int pt) const; //only called on fat points

    // A set is 'minimal' if there is no point for which all lines have weight > 1
    bool isMinimal() const ;

    // Compute and store the invariant values of all points of the set
    void computeInvariants () ;

    void printWeights() const;

private:
    // Compute invariant for the given FAT point. No need to compute invariants for the other points
    int computeInvariant (int pt) const;

};

/////////////////////////////////////////
// DEFINITIONS
////////////////////////////////////////

template<int Q>
void ThreeWeights<Q>::add(int pt) { // 0 <= pt < F
    set.add(pt);
    // fat line through origin
    int m = pt % K;
    weights[F + m] ++;
    // Q fat lines through pt; one of them is the special fat line through this point (consisting of the three thin lines forming the triangle through the 3 thin points this fat points is made up of)
    // This special line will be iterated over twice, and hence will have its weight increased twice, as is desired.
    for (int el: Semi<Q>::D) {
        int ln = (pt + 3 * F - el) % F;
        weights[ln]++;
    }
}

template<int Q>
void ThreeWeights<Q>::addInfty(int pt) {// we assume that this fat pt is in 0,...,(Q-2)/3 (or: 0,..,K-1) as the q+1 pts at inf are partitioned into (q+1)/3 (or: K) sets of size 3.
    // The partition works as follows: The automorphism of order 3 working on the plane is just i -> i+(q²-1)/3; for q=5 this is i->i+8, for q=11 this is i->i+40.
    // i-> i+a with a = (q²-1)/3 mod (q+1) is the action of this automorphism on the pts at inf (Q_0,Q_1,...,Q_q); for q=5 this is i->i+2 and hence Q_0 Q_2 Q_4 form an orbit at inf, Q_1 Q_3 Q_5 is the other one.
    // for q=11, this is i->i+4 and hence the orbits at inf are: 0,4,8 ; 1,5,9 ; 2,6,10 ; 3,7,11 (these are the indices j for Q_j).
    // The same holds for the lines through the origin; they are partitioned into orbits with the exact same indices.
    set.addInfty(pt);
    //line at infinity
    weights[V - 1]+=3;
    //fat line through origin and this fat pt at inf
    weights[F + pt]++;
    // q-1 fat lines through pt
    for (int i = 0; i < Q-1; ++i) {
        weights[pt + i * K]++;
    }
}

template<int Q>
void ThreeWeights<Q>::addOrigin() {
    set.addOrigin();
    for (int i = 0; i < K; ++i) {
        weights[F + i]++;
    }
}

template<int Q>
void ThreeWeights<Q>::remove(int pt) { //reverse what is done in add
    int m = pt % K;
    weights[F + m] --;
    for (int el : Semi<Q>::D) {
        int ln = (pt + 3 * F - el) % F;
        weights[ln]--;
    }
    set.remove(pt);
}

template<int Q>
void ThreeWeights<Q>::removeInfty(int pt) {// undoing what is done in addInfty; documentation can be found there!
    weights[V - 1]-=3;
    weights[F + pt]--;
    for (int i = 0; i < Q-1; ++i) {
        weights[pt + i * K]--;
    }
    set.removeInfty(pt);
}

template<int Q>
void ThreeWeights<Q>::removeOrigin() {
    for (int i = 0; i < K; ++i) {
        weights[F + i]--;
    }
    set.removeOrigin();
}

template<int Q>
bool ThreeWeights<Q>::isSuperfluous(int pt) const {
    // fat line through origin
    int m = pt % K;
    if (weights[F + m] == 0) { return false; }
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 3 * F - el) % F;
        if (weights[ln] == 0) { return false; }
    }
    return true;
}

template<int Q>
bool ThreeWeights<Q>::isRedundantFat(int pt) const {
    // fat line through origin
    int m = pt % K;
    if(weights[F+m]<2){return false;}
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 3 * F - el) % F;
        if (weights[ln] < 2) { return false; }
    }
    return true;
}

template<int Q>
bool ThreeWeights<Q>::isRedundantInfty(int pt) const {
    if (weights[V - 1] < 2 || weights[F + pt] < 2) { return false; } // line at inf and fat line through origin. First check is useless
    int ln = pt % K; // ln is the fat line in 0,...,Q that goes through pt
    for (int i = 0; i < Q-1; ++i) {
        if (weights[pt + i * K] < 2) {
            return false;
        }
    }
    return true;
}

template<int Q>
bool ThreeWeights<Q>::isRedundantOrigin() const {
    for (int i = 0; i < K; ++i) {
        if (weights[F + i] < 2) {
            return false;
        }
    }
    return true;
}

template<int Q>
int ThreeWeights<Q>::computeInvariant(int pt) const {
    int total = 0;
    // fat line through origin
    int m = pt % K;
    total += hashedWeight(F + m);
    // Q fat lines through pt
    for (int el : Semi<Q>::D) {
        int ln = (pt + 3 * F - el) % F;
        total += hashedWeight(ln);
    }
    return total;
}

template<int Q>
bool ThreeWeights<Q>::isMinimal() const {
    // check whether any of the fat points is redundant
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt) && isRedundantFat(pt)) {
            return false;
        }
    }
    // check whether any of the points at infinity is redundant
    for (int pt = 0; pt < K; ++pt) {
        if (set.contains(F + pt) && isRedundantInfty(pt)) {
            return false;
        }
    }
    return !(set.contains(V-1) && isRedundantOrigin());
}

// TODO almost same as in Singer Weights
template<int Q>
void ThreeWeights<Q>::computeInvariants() {
    for (int pt = 0; pt < F; pt++) {
        if (set.contains(pt)) {
            this->invariants[pt] = computeInvariant(pt);
        }
    }
}

template<int Q>
void ThreeWeights<Q>::printWeights() const {
    //printing weights
    std::cout << "WEIGHTS ";
    for (int j = 0; j <= Q+1; ++j) {
        std::cout << j << ": ";
        for (int i = 0; i < V; ++i) {
            if (this->weights[i] == j) {
                std::cout << i << "-";
            }
        }
        std::cout << std::endl;
    }
}

#endif //BSETSNONTRIV_THREEWEIGHTS_HPP
